@@ -7,7 +7,8 @@ if sys.version < '2.4':
     from sets import ImmutableSet as frozenset
 
 from routes import request_config
-from routes.util import controller_scan, RouteException
+from routes.lru import LRUCache
+from routes.util import controller_scan, MatchException, RoutesException
 from routes.route import Route
 
 
@@ -20,6 +21,38 @@ def strip_slashes(name):
     return name
 
 
+class SubMapper(object):
+    """Partial mapper for use with_options"""
+    def __init__(self, obj, **kwargs):
+        self.kwargs = kwargs
+        self.obj = obj
+        
+    def connect(self, *args, **kwargs):
+        newkargs = {}
+        newargs = args
+        for key in self.kwargs:
+            if key == 'path_prefix':
+                if len(args > 1):
+                    newargs = (args[0], self.kwargs[key] + args[1])
+                else:
+                    newargs = (self.kwargs[key] + args[0],)
+            elif key in kwargs:
+                newkargs[key] = self.kwargs[key] + kwargs[key]
+            else:
+                newkargs[key] = self.kwargs[key]
+        for key in kwargs:
+            if key not in self.kwargs:
+                newkargs[key] = kwargs[key]
+        return self.obj.connect(*newargs, **newkargs)
+
+    # Provided for those who prefer using the 'with' syntax in Python 2.5+
+    def __enter__(self):
+        return self
+    
+    def __exit__(self, type, value, tb):
+        pass
+
+
 class Mapper(object):
     """Mapper handles URL generation and URL recognition in a web
     application.
@@ -44,6 +77,10 @@ class Mapper(object):
             ``directory`` keyword arg is present, it will be passed
             into the function during its call. This option defaults to
             a function that will scan a directory for controllers.
+            
+            Alternatively, a list of controllers or None can be passed
+            in which are assumed to be the definitive list of
+            controller names valid when matching 'controller'.
         
         ``directory``
             Passed into controller_scan for the directory to scan. It
@@ -98,7 +135,7 @@ class Mapper(object):
         self.matchlist = []
         self.maxkeys = {}
         self.minkeys = {}
-        self.urlcache = {}
+        self.urlcache = LRUCache(1600)
         self._created_regs = False
         self._created_gens = False
         self.prefix = None
@@ -124,13 +161,88 @@ class Mapper(object):
             config.mapper = self
     
     def _envget(self):
-        return getattr(self.req_data, 'environ', None)
+        try:
+            return self.req_data.environ
+        except AttributeError:
+            return None
     def _envset(self, env):
         self.req_data.environ = env
     def _envdel(self):
         del self.req_data.environ
     environ = property(_envget, _envset, _envdel)
-
+    
+    def submapper(self, **kargs):
+        """Create a partial version of the Mapper with the designated
+        options set
+        
+        This results in a :class:`routes.mapper.SubMapper` object.
+        
+        Only keyword arguments can be saved for use with the submapper
+        and only a 'connect' method is present on the submapper.
+        
+        If keyword arguments provided to this method also exist in the
+        keyword arguments provided to the submapper, their values will
+        be merged with the saved options going first.
+        
+        In addition to :class:`routes.route.Route` arguments, submapper
+        can also take a ``path_prefix`` argument which will be
+        prepended to the path of all routes that are connected.
+        
+        Example::
+            
+            >>> map = Mapper(controller_scan=None)
+            >>> map.connect('home', '/', controller='home', action='splash')
+            >>> map.matchlist[0].name == 'home'
+            True
+            >>> m = map.submapper(controller='home')
+            >>> m.connect('index', '/index', action='index')
+            >>> map.matchlist[1].name == 'index'
+            True
+            >>> map.matchlist[1].defaults['controller'] == 'home'
+            True
+        
+        """
+        return SubMapper(self, **kargs)
+    
+    def extend(self, routes, path_prefix=''):
+        """Extends the mapper routes with a list of Route objects
+        
+        If a path_prefix is provided, all the routes will have their
+        path prepended with the path_prefix.
+        
+        Example::
+            
+            >>> map = Mapper(controller_scan=None)
+            >>> map.connect('home', '/', controller='home', action='splash')
+            >>> map.matchlist[0].name == 'home'
+            True
+            >>> routes = [Route('index', '/index.htm', controller='home',
+            ...                 action='index')]
+            >>> map.extend(routes)
+            >>> len(map.matchlist) == 2
+            True
+            >>> map.extend(routes, path_prefix='/subapp')
+            >>> len(map.matchlist) == 3
+            True
+            >>> map.matchlist[2].routepath == 'subapp/index.htm'
+            True
+        
+        .. note::
+            
+            This function does not merely extend the mapper with the
+            given list of routes, it actually creates new routes with
+            identical calling arguments.
+        
+        """
+        for route in routes:
+            if path_prefix and route.minimization:
+                routepath = '/'.join([path_prefix, route.routepath])
+            elif path_prefix:
+                routepath = path_prefix + route.routepath
+            else:
+                routepath = route.routepath
+            self.connect(route.name, routepath, **route._kargs)
+                
     def connect(self, *args, **kargs):
         """Create and connect a new Route to the Mapper.
         
@@ -151,7 +263,8 @@ class Mapper(object):
         routename = None
         if len(args) > 1:
             routename = args[0]
-            args = args[1:]
+        else:
+            args = (None,) + args
         if '_explicit' not in kargs:
             kargs['_explicit'] = self.explicit
         if '_minimize' not in kargs:
@@ -165,9 +278,12 @@ class Mapper(object):
             route.encoding = self.encoding
             route.decode_errors = self.decode_errors
         
-        self.matchlist.append(route)
+        if not route.static:
+            self.matchlist.append(route)
+        
         if routename:
             self._routenames[routename] = route
+            route.name = routename
         if route.static:
             return
         exists = False
@@ -237,8 +353,12 @@ class Mapper(object):
         if clist is None:
             if self.directory:
                 clist = self.controller_scan(self.directory)
-            else:
+            elif callable(self.controller_scan):
                 clist = self.controller_scan()
+            elif not self.controller_scan:
+                clist = []
+            else:
+                clist = self.controller_scan
             
         for key, val in self.maxkeys.iteritems():
             for route in val:
@@ -263,7 +383,7 @@ class Mapper(object):
         if not self._created_regs and self.controller_scan:
             self.create_regs()
         elif not self._created_regs:
-            raise RouteException("You must generate the regular expressions"
+            raise RoutesException("You must generate the regular expressions"
                                  " before matching.")
         
         if self.always_scan:
@@ -277,16 +397,21 @@ class Mapper(object):
                     url = '/'
             else:
                 return (None, None, matchlog)
+        environ = self.environ
+        sub_domains = self.sub_domains
+        sub_domains_ignore = self.sub_domains_ignore
+        domain_match = self.domain_match
+        debug = self.debug
         for route in self.matchlist:
             if route.static:
-                if self.debug:
+                if debug:
                     matchlog.append(dict(route=route, static=True))
                 continue
-            match = route.match(url, self.environ, self.sub_domains, 
-                self.sub_domains_ignore, self.domain_match)
-            if self.debug:
+            match = route.match(url, environ, sub_domains, sub_domains_ignore,
+                                domain_match)
+            if debug:
                 matchlog.append(dict(route=route, regexp=bool(match)))
-            if match:
+            if isinstance(match, dict) or match:
                 return (match, route, matchlog)
         return (None, None, matchlog)
     
@@ -301,13 +426,13 @@ class Mapper(object):
         
         """
         if not url:
-            raise RouteException('No URL provided, the minimum URL necessary'
+            raise RoutesException('No URL provided, the minimum URL necessary'
                                  ' to match is "/".')
         
         result = self._match(url)
         if self.debug:
             return result[0], result[1], result[2]
-        if result[0]:
+        if isinstance(result[0], dict) or result[0]:
             return result[0]
         return None
     
@@ -325,7 +450,7 @@ class Mapper(object):
         result = self._match(url)
         if self.debug:
             return result[0], result[1], result[2]
-        if result[0]:
+        if isinstance(result[0], dict) or result[0]:
             return result[0], result[1]
         return None
     
@@ -345,7 +470,7 @@ class Mapper(object):
         
         if self.append_slash:
             kargs['_append_slash'] = True
-                
+        
         if not self.explicit:
             if 'controller' not in kargs:
                 kargs['controller'] = 'content'
@@ -373,14 +498,14 @@ class Mapper(object):
                 if key in self.urlcache:
                     return self.urlcache[key]
         
-        actionlist = self._gendict.get(controller) or self._gendict.get('*')
-        if not actionlist:
+        actionlist = self._gendict.get(controller) or self._gendict.get('*', {})
+        if not actionlist and not args:
             return None
         (keylist, sortcache) = actionlist.get(action) or \
-                               actionlist.get('*', (None, None))
-        if not keylist:
+                               actionlist.get('*', (None, {}))
+        if not keylist and not args:
             return None
-        
+
         keys = frozenset(kargs.keys())
         cacheset = False
         cachekey = unicode(keys)
@@ -445,6 +570,10 @@ class Mapper(object):
                 kval = kargs.get(key)
                 if not kval:
                     continue
+                if isinstance(kval, str):
+                    kval = kval.decode(self.encoding)
+                else:
+                    kval = unicode(kval)
                 if kval != route.defaults[key]:
                     fail = True
                     break
@@ -454,8 +583,9 @@ class Mapper(object):
             if path:
                 if self.prefix:
                     path = self.prefix + path
+                external_static = route.static and route.external
                 if self.environ and self.environ.get('SCRIPT_NAME', '') != ''\
-                    and not route.absolute:
+                    and not route.absolute and not external_static:
                     path = self.environ['SCRIPT_NAME'] + path
                     key = cache_key_script_name
                 else:
@@ -658,6 +788,7 @@ class Mapper(object):
             '_member_name': member_name,
             '_collection_name': collection_name,
             '_parent_resource': parent_resource,
+            '_filter': kwargs.get('_filter')
         }
         
         def requirements_for(meth):
@@ -737,3 +868,54 @@ class Mapper(object):
         self.connect("formatted_" + name_prefix + member_name, 
                      member_path + ".:(format)", **route_options)
         self.connect(name_prefix + member_name, member_path, **route_options)
+    
+    def redirect(self, match_path, destination_path, *args, **kwargs):
+        """Add a redirect route to the mapper
+        
+        Redirect routes bypass the wrapped WSGI application and instead
+        result in a redirect being issued by the RoutesMiddleware. As
+        such, this method is only meaningful when using
+        RoutesMiddleware.
+        
+        By default, a 302 Found status code is used, this can be
+        changed by providing a ``_redirect_code`` keyword argument
+        which will then be used instead. Note that the entire status
+        code string needs to be present.
+        
+        When using keyword arguments, all arguments that apply to
+        matching will be used for the match, while generation specific
+        options will be used during generation. Thus all options
+        normally available to connected Routes may be used with
+        redirect routes as well.
+        
+        Example::
+            
+            map = Mapper()
+            map.redirect('/legacyapp/archives/{url:.*}, '/archives/{url})
+            map.redirect('/home/index', '/', _redirect_code='301 Moved Permanently')
+        
+        """
+        both_args = ['_encoding', '_explicit', '_minimize']
+        gen_args = ['_filter']
+        
+        status_code = kwargs.pop('_redirect_code', '302 Found')
+        gen_dict, match_dict = {}, {}
+        
+        # Create the dict of args for the generation route
+        for key in both_args + gen_args:
+            if key in kwargs:
+                gen_dict = kwargs[key]
+        gen_dict['_static'] = True
+        
+        # Create the dict of args for the matching route
+        for key in kwargs:
+            if key not in gen_args:
+                match_dict[key] = kwargs[key]
+        
+        self.connect(match_path, **match_dict)
+        match_route = self.matchlist[-1]
+        
+        self.connect('_redirect_%s' % id(match_route), destination_path,
+                     **gen_dict)
+        match_route.redirect = True
+        match_route.redirect_status = status_code
